/*--------------------------------------*/
/*		3D Studio		*/
/*  Image Processing Function Interface	*/
/*	  by Tom Hudson 2/5/92		*/
/*    Revised 8/14/92 for Version 2.01  */
/*--------------------------------------*/

/* Change the next line to "#define DBGIXP" for debugging */

#define xxDBGIXP

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <pltypes.h>
#include "ixp.h"

typedef unsigned int uint;
typedef unsigned short ushort;
typedef unsigned long ulong;

char gp_buffer[256];

static char *state;
static DlgEntry *cdialog;
static int state_size;

static FrameParm fparm;
static FrameInfo finfo;
static FieldInfo fldinfo;

#define CFGTITLE 0
#define CFGONOFF 1
#define CFGYESNO 2
#define CFGRADIO 3
#define CFGLFLOAT 4
#define CFGFLOAT 5
#define CFGLINT 6
#define CFGINT 7
#define CFGFILENAME 8
#define CFGSTRING 9
#define CFGSLIDER 10
#define CFGVAR 11
#define CFGCOLOR 11

static int FindDlgType(char *s) {
	if (s[0]=='-') s++;
	if (strncmp(s,"TITLE",5)==0) return(CFGTITLE);
	if (strncmp(s,"ON-OFF",5)==0) return(CFGONOFF);
	if (strncmp(s,"YES-NO",5)==0) return(CFGYESNO);
	if (strncmp(s,"RADIO",5)==0) return(CFGRADIO);
	if (strncmp(s,"LFLOAT",6)==0) return(CFGLFLOAT);
	if (strncmp(s,"FLOAT",5)==0) return(CFGFLOAT);
	if (strncmp(s,"LINT",4)==0) return(CFGLINT);
	if (strncmp(s,"INT",3)==0) return(CFGINT);
	if (strncmp(s,"SLIDER",6)==0) return(CFGSLIDER);
	if (strncmp(s,"COLOR",5)==0) return(CFGCOLOR);
	if (strncmp(s,"FILENAME",8)==0) return(CFGFILENAME);
	if (strncmp(s,"STRING",6)==0) return(CFGSTRING);
	return(0);
	}

static int DlgHasVar(char *s) {
	if (s[0]=='-') s++;
	if (strncmp(s,"TITLE",5)==0) return(0);
	else return(1);
	}

typedef struct{
	ulong magic;	/* Magic number */
	int opcode;		/* What .EXP wants to do in 3D Studio */
	int usercode;	/* What .EXP wants to do when it gets back */
	int status;		/* How it all came out */
	} BufHdr;


typedef struct{
	unsigned long size;
	ulong data;
	} IXPcfg;

static int bufsize=sizeof(EXPbuf);


/* Pointers to bitmaps living in 3D Studio's memory	*/
/* These are "_far" pointers -- 48 bits worth		*/

static EXPbuf _far *fexpbuf;	/* Comm buffer struct ptr */
static void _far *fstate;

/*-------------------------------------------------*/

/* Local Data */
static EXPbuf localbuf;
static Report report;

#define ixd localbuf.data.ixp

static int dialog_ix, nvar, num_dialog_items;
static int version=200;

long xpmagic,xpmainseg,xpbuffer;

static InitIxp() {
	int bufs2;
	state = ClientGetState(&state_size);
	bufsize = sizeof(EXPbuf);
	bufs2 = state_size+sizeof(BufHdr)+4;
	if (bufs2>bufsize) bufsize = bufs2;
	}

main(void)	{
	int type,phase,i,k;
	ulong *pl;
	OVL val;

	/* Check magic number */
	if(xpmagic!=IXP_MAGIC)	{
	bad_magic:
		localbuf.status=0;
		return;
		}

	/* Copy comm buffer data to our structure */
	FP_SET(fexpbuf,xpbuffer,xpmainseg);
	far_to_near(&localbuf,fexpbuf,bufsize);

	/* Double-check its magic number */
	
	if(localbuf.magic!=IXP_MAGIC)
		goto bad_magic;

	if(localbuf.ver.magic==VER_MAGIC)
		version=localbuf.ver.version;
/*printf("3DS Version= %d.%02d\n",version/100,version-(version/100)*100);*/

#if 0
	printf(" usercode = %X , bufsize = %d\n",localbuf.usercode,bufsize);
	getchar();
#endif

	switch(localbuf.usercode) {
		case EXP_STARTUP:		/* Initial startup */
			InitIxp();
			ClientStartup();
			break;

		case IXP_PROC_IMAGE:
#define di localbuf.data.ixp
			fparm.dev_width=di.width;
			fparm.dev_height=di.height;
			fparm.dev_aspect=di.aspect;
			fparm.curfield=di.curfield;
			fparm.totfields=di.totfields;

			/*
			Get the pointers in 3D Studio's memory where the bitmaps are located.
			*/

			FP_SET(fparm.inpix,di.imagein,xpmainseg);
			FP_SET(fparm.outpix,di.imageout,xpmainseg);
			if(di.alphin)
				{
				fparm.gotalpha=1;
				FP_SET(fparm.inalph,di.alphin,xpmainseg);
				}
			else
				{
				fparm.gotalpha=0;
				fparm.inalph=(uchar _far *)NULL;
				}

			/* Ask 3D Studio about the current view */

			localbuf.opcode=EXP_FRAME_INF;
			localbuf.status=1;
			localbuf.usercode=0x0005;
			break;
		case 0x0005:
			if(localbuf.status==0)
				{
				localbuf.opcode=localbuf.usercode=EXP_TERMINATE;
				localbuf.status=0;
				break;
				}

			memcpy(&finfo,&localbuf.data.finfo,sizeof(FrameInfo));

			if(version==200)
				{
				fldinfo.using_fields=0;
				goto no_fieldinfo;
				}

			/* Ask 3D Studio about the field info */

			localbuf.opcode=EXP_FIELD_INF;
			localbuf.status=1;
			localbuf.usercode=0x0004;
			break;
		case 0x0004:
			if(localbuf.status==0)
				{
				localbuf.opcode=localbuf.usercode=EXP_TERMINATE;
				localbuf.status=0;
				break;
				}
			memcpy(&fldinfo,&localbuf.data.fldinfo,sizeof(FieldInfo));

			no_fieldinfo:
			phase=0;

			recall:
			switch(ClientProcImage(&fparm,&finfo,&fldinfo,phase,&report)) /* what we're here for */
				{
				case 0:
					localbuf.usercode=EXP_TERMINATE;
					localbuf.status=0;
					break;
				case 1:
					localbuf.usercode=EXP_TERMINATE;
					localbuf.status=1;
					break;
				case 2:
					localbuf.opcode=EXP_PROGRESS;
					localbuf.usercode=0x0010;
#define prog localbuf.data.progress
					strcpy(prog.title,report.title);
					prog.max=report.max;
					prog.current=report.current;
#undef prog
					localbuf.status=1;
					break;
				}
			break;
		case 0x0010:	/* After status report */
			phase=report.recall_state;
			goto recall;

		case EXP_GET_CONFIG:
			state = ClientGetState(&state_size);
  	#define pd localbuf.data.cfig
			if(pd.size!=state_size) {
				ClientResetState();
				}
			else 	{
 				long *p, *q;
				p = (long *)&pd.data;
				q = (long *)state;
				if (p[0]!=q[0]) 
					ClientResetState();
				else 				
					memcpy(state,&pd.data,state_size);
				}
	#undef pd
			cdialog = ClientDialog(0);
			if(cdialog->s==NULL) {
				goto var_done;
				}
			localbuf.opcode=EXP_INIT_DIALOG;
			localbuf.usercode=0x0100;
			localbuf.status=1;
			break;
	
		case 0x0100:
			dialog_ix=0;
	
		case 0x0110:
			cdialog = ClientDialog(dialog_ix);
			if(cdialog->s==NULL)
				goto dia_done;
			strcpy(localbuf.data.string,cdialog->s);
			localbuf.opcode=EXP_ADD_DIALOG;
			localbuf.usercode=0x0110;
			localbuf.status=1;
			dialog_ix++;
			break;
	
			dia_done:
			num_dialog_items = dialog_ix;

			dialog_ix=0;
			nvar = 0;
	
		case 0x0120:
			for ( ; dialog_ix<num_dialog_items ; dialog_ix++) {
				cdialog = ClientDialog(dialog_ix);
				if (DlgHasVar(cdialog->s)) break;
				}
			if (dialog_ix>=num_dialog_items) 
				goto var_done;
			val.ul = ClientGetStateVar(cdialog->id);
			type = FindDlgType(cdialog->s);
			switch(type) {
				case CFGLINT: case CFGINT:
					sprintf(localbuf.data.string,"VARIABLE=\"%d\"",val.i);
					nvar++;
					break;
				case CFGONOFF: case CFGYESNO:
				case CFGRADIO: case CFGSLIDER:
					sprintf(localbuf.data.string,"VARIABLE=%d",val.i);
					nvar++;
					break;
				case CFGCOLOR:
					sprintf(localbuf.data.string,"VARIABLE=%d,%d,%d",val.c.r,val.c.g,val.c.b);
					nvar++;
					break;
				case CFGFLOAT: case CFGLFLOAT:
					sprintf(localbuf.data.string,"VARIABLE=\"%f\"",val.f);
					nvar++;
					break;
				case CFGSTRING: case CFGFILENAME:
					sprintf(localbuf.data.string,"VARIABLE=\"%s\"",val.s);
					nvar++;
					break;
				}
#ifdef DBGIXP
		  		printf("  data.string =  %s \n",localbuf.data.string);
#endif
			localbuf.opcode=EXP_ADD_VAR;
			localbuf.usercode=0x0120;
			localbuf.status=1;
			dialog_ix++;
			break;
	
			var_done:
			localbuf.opcode=EXP_DO_DIALOG;
			localbuf.usercode=0x0130;
			localbuf.status=1;
			break;
		case 0x0130:
			if(localbuf.status==0)	/* Cancelled */
				{
				localbuf.opcode=localbuf.usercode=EXP_TERMINATE;
				localbuf.status=0;
				break;
				}
			pl = (ulong *)&localbuf.data;
			k = 0;
			for (i=0; ; i++) {
				cdialog = ClientDialog(i);
				if (cdialog->s==NULL) break;
#ifdef DBGIXP
				printf("  DLG[%d]: %s (k=%d)\n",i,cdialog->s,k);
#endif
				if (DlgHasVar(cdialog->s)) 
					{
					ClientSetStateVar(cdialog->id,&pl[k]);
					k+=ClientVarSize(cdialog->id);
					}
				}

			/* done with the dialog session -- send the state back. */
			localbuf.data.cfig.size = state_size;
			memcpy(&localbuf.data.cfig.data,state,state_size);
			localbuf.opcode=EXP_PROC_DATA;
			localbuf.usercode=EXP_TERMINATE;
			localbuf.status=1;
			break;

		case EXP_PUT_CONFIG:
	#define pd localbuf.data.cfig
			if(pd.size)	{
				long *p, *q;
				if(pd.size!=state_size) {
					goto re_state;
					}
				p = (long *)&pd.data;
				q = (long *)state;
				if (p[0]!=q[0]) {
					goto re_state;
					}
				memcpy(state,&pd.data,state_size);			
				}
			else {
				re_state:
				ClientResetState();
				}
	#undef pd
			localbuf.opcode = EXP_CONFIGURED;
			localbuf.status = 1;
			break;

		case EXP_TERMINATE:
#ifdef DBGIXP
			printf("ClientTerminate \n");		
#endif
			ClientTerminate();
			localbuf.opcode=EXP_TERMINATE;	/* Terminate! */
			localbuf.status=1;	/* OK status */
			break;

		default:
			printf("ERROR IN IXP.C: %X\n",localbuf.usercode);

			term_err:
			ClientTerminate();
			localbuf.opcode=EXP_TERMINATE;	/* Terminate! */
			localbuf.status=0;	/* ERROR status */
			break;
		}
	
	near_to_far(fexpbuf,&localbuf,bufsize);
	return;
	}

/* Write error information out to a file */

error(char *s1,char *s2,char *s3,char *s4,char *s5) {
	FILE *stream;

	if((stream=fopen("rend.err","w"))==NULL)
	 return;
	fprintf(stream,"%s error:\n",ClientName());
	if(s1!=NULL)
	 fprintf(stream,"%s\n",s1);
	if(s2!=NULL)
	 fprintf(stream,"%s\n",s2);
	if(s3!=NULL)
	 fprintf(stream,"%s\n",s3);
	if(s4!=NULL)
	 fprintf(stream,"%s\n",s4);
	if(s5!=NULL)
	 fprintf(stream,"%s\n",s5);
	fclose(stream);
	}

int
studio_version()
{
return(version);
}
