/*       3D Studio's              */
/*  Ripple-object Process ver 1.1 */
/*  by Tom Hudson 11/14/91        */
/* (Modifed for 2.01 burst mode.) */

/* PXP process raw client routine core code */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "pxp.h"
#include "trig.h"

extern abswarp(float *x,float *y,float *z,float cx,float cy);
extern relwarp(float *x,float *y,float *z,float cx,float cy,float cz);

/* Types of waves this thing can make... */

#define XY 0
#define XZ 1
#define YZ 2
#define XYZ 3

/* Absolute or Relative waves */

#define ABSOLUTE 0
#define RELATIVE 1

/* Variable definitions */

#define AXIS 1
#define TYPE 2
#define PERIOD 3
#define WAVELENGTH 4
#define AMP 5
#define MORPHS 6
#define HIDE 7
#define PREFIX 8

/************************/

/* Dialog description */

DlgEntry cdialog[]={
	0,"TITLE=\"Autodesk 3D Studio Radial Wave Generator\"",
	0,"TITLE=\"by Tom Hudson, Version 1.1\"",
	0,"TITLE=\"\"",
	AXIS,"RADIO=\"Amplitude axis:\",\"Top\",\"Front\",\"Right\",\"All\"",
	TYPE,"RADIO=\"Ripple type:\",\"Absolute\",\"Relative\"",
	0,"TITLE=\"\"",
	PERIOD,"LINT=\"Period:\",0,500,3",
	WAVELENGTH,"-LFLOAT=\"Wavelength:\",0.001,99999,8",
	AMP,"-LFLOAT=\" Amplitude:\",0.0,99999,8",
	0,"TITLE=\"\"",
	MORPHS,"LINT=\"# of morphs:\",1,999,3",
	HIDE,"-YES-NO=\"Hide >0?\"",
	PREFIX,"-STRING=\"Name prefix:\",7,7",
	0,"TITLE=\"\"",
	0,NULL
	};

/* Version test value */

#define VERSION 0x71B0
	
typedef struct {
	ulong version;
	int axis;
	int type;
	int period;
	float wavelength;
	float amp;
	int morphs;
	int hide;
	char prefix[8];
	} State;

/* the "state" struct MUST start with a "ulong" which is the version#,
	to prevent using data from old versions of this program.
	This verification is performed automatically. */

static State init_state = { VERSION,XY,ABSOLUTE,10,10.0,1.0,10,1,"RIPPLE" };
static State state = { VERSION,XY,ABSOLUTE,10,10.0,1.0,10,1,"RIPPLE" };

static float centerx,centery,centerz;
static int verts,faces,tverts,ix;
static int index,vertex,face;
static int morphnum;

/* Declare burst mode arrays */
static VData *vdata=NULL;

static FData *fdata=NULL;

/*----------------------------------------------------------------*/

void ClientSetStateVar(int id, void *ptr) {
	OVL o;
	ulong *ul;
	char *s;

	ul=(ulong *)ptr;
	s=(char *)ptr;
	o.ul = *ul;
	switch(id) {
		case AXIS: state.axis=o.i; break;
		case TYPE: state.type=o.i; break;
		case PERIOD: state.period=o.i; break;
		case WAVELENGTH: state.wavelength=o.f; break;
		case AMP: state.amp=o.f; break;
		case MORPHS: state.morphs=o.i; break;
		case HIDE: state.hide=o.i; break;
		case PREFIX: strcpy(state.prefix,s); break;
		}
	}

ulong ClientGetStateVar(int id) {
	OVL o;
	switch(id) {
		case AXIS: o.i=state.axis; break;
		case TYPE: o.i=state.type; break;
		case PERIOD: o.i=state.period; break;
		case WAVELENGTH: o.f=state.wavelength; break;
		case AMP: o.f=state.amp; break;
		case MORPHS: o.i=state.morphs; break;
		case HIDE: o.i=state.hide; break;
		case PREFIX: o.s=state.prefix; break;
		}
	return(o.ul);
	}

int ClientVarSize(int id)
	{
	switch(id)
		{
		case PREFIX: return(2);
		default:
			return(1);
		}
	}

char  *ClientGetState(int *size) {
	*size = sizeof(State);
	return((char *)&state);
	}

void ClientResetState() {   
	state = init_state; 
	}

void ClientStartup(EXPbuf *buf) {
	buf->opcode=EXP_NOP;    /* NO-OP */
	buf->usercode=0x0200;
	}

/* User routines -- use user codes to process your data */

void ClientUserCode(EXPbuf *buf)
	{
	switch(buf->usercode)
		{
		case 0x0200:
			/* See if there will be name conflicts */

			index=0;

			conflict_loop:
#define ob buf->data.object
			sprintf(ob.name,"%s%03d",state.prefix,index);
#undef ob
			buf->opcode=EXP_OBJ_EXIST;
			buf->usercode=0x0205;
			break;
		case 0x0205:
			if(buf->status)
				{
				strcpy(buf->data.string,"Name conflict -- use a unique prefix");
				buf->opcode=EXP_CONTINUE;
				buf->usercode=EXP_TERMINATE;
				break;
				}
			index++;
			if(index<state.morphs)
				goto conflict_loop;

			buf->opcode=EXP_PICK_OBJECT;    /* Studio does this */
			buf->usercode=0x0210;       /* Do this next time */
			strcpy(buf->data.string,"Pick a mesh object for center");
			break;
  
		case 0x0210:
			if(buf->status==0)
				goto terminate;
			buf->opcode=EXP_OBJ_EXTENTS;
			buf->usercode=0x0212;
			break;

		case 0x0212:
			if(buf->status==0)
				goto terminate;
#define oe buf->data.exts
			centerx=oe.centerx;
			centery=oe.centery;
			centerz=oe.centerz;
#undef oe
			buf->opcode=EXP_PICK_OBJECT;    /* Studio does this */
			buf->usercode=0x0214;       /* Do this next time */
			strcpy(buf->data.string,"Pick a mesh object to ripple");
			break;
  
		case 0x0214:
			if(buf->status==0)
				goto terminate;

			/* These vert and face counts determine size of burst mode 
			data arrays */

#define co buf->data.object
			verts=co.verts;
			tverts=co.tverts;
			faces=co.faces;
#undef co

			morphnum=0;

			/* Allocate RAM for object work buffers for burst mode */
			if(studio_version()!=200)
				{
				if((vdata=(VData *)malloc(verts*sizeof(VData)))==NULL)
					{
					noram:
					buf->opcode=EXP_CONTINUE;
					buf->usercode=EXP_TERMINATE;
					strcpy(buf->data.string,"No RAM for object workspace");
					return;
					}
				if((fdata=(FData *)malloc(faces*sizeof(FData)))==NULL)
					goto noram;
				}

			morph_loop:

			/* Set up for an exact copy of this object */

			buf->opcode=EXP_READY_OBJ;
			buf->usercode=0x0215;
#define ob buf->data.object
			ob.verts=verts;
			ob.faces=faces;
			ob.tverts=tverts;
#undef ob
			break;
		case 0x0215:
			if(buf->status==0)
				goto terminate;
			buf->opcode=EXP_PROMPT;
			sprintf(buf->data.string,
				"Creating Ripple Object %d...",morphnum);
			buf->usercode=0x0220;
			break;

			/* Copy face structure */
/* check 3DS version, and use burst mode if it's not version 2.0 */

		case 0x0220:
			if(studio_version()!=200)
				goto burst_mode;

			face=0;

			face_loop:
			buf->opcode=EXP_GET_FACE;
			buf->usercode=0x0225;
#define fc buf->data.face
			fc.face=face;
#undef fc
			break;

		case 0x0225:
			buf->opcode=EXP_PUT_FACE;
			buf->usercode=0x0230;
#define fc buf->data.face
			fc.face=face;
#undef fc
			break;

		case 0x0230:
			face++;
			if(face<faces)
				goto face_loop;

			/* Faces copied, now copy verts with warping */

			vertex=0;

			vert_loop:
			buf->opcode=EXP_GET_VERTEX;
			buf->usercode=0x0235;
#define vt buf->data.vert
			vt.vertex=vertex;
#undef vt
			break;

		case 0x0235:

			/* Got vertex, now modify XYZ based on parameters */

#define vt buf->data.vert
			if(state.type==ABSOLUTE)
				{
				switch(state.axis)
					{
					case XY:
						abswarp(&vt.x,&vt.y,&vt.z,centerx,centery);
						break;
					case XZ:
						abswarp(&vt.x,&vt.z,&vt.y,centerx,centerz);
						break;
					case YZ:
						abswarp(&vt.y,&vt.z,&vt.x,centery,centerz);
						break;
					case XYZ:
						abswarp(&vt.x,&vt.y,&vt.z,centerx,centery);
						abswarp(&vt.x,&vt.z,&vt.y,centerx,centerz);
						abswarp(&vt.y,&vt.z,&vt.x,centery,centerz);
						break;
					}
				}
			else
				relwarp(&vt.x,&vt.y,&vt.z,centerx,centery,centerz);

			buf->opcode=EXP_PUT_VERTEX;
			buf->usercode=0x0240;
			vt.vertex=vertex;
#undef vt
			break;

		case 0x0240:
			vertex++;
			if(vertex<verts)
				goto vert_loop;
			goto create_it;


			burst_mode:
			/* Copy vertex structure into RAM using burst mode command
			EXP_GET_VLIST. */

			buf->opcode=EXP_GET_VLIST;
			buf->usercode=0x0250;
#define cv buf->data.vlist
			cv.start=0;
			cv.count=verts;
			cv.data=(_far VData *)vdata;
#undef cv
			break;

			/* Copy face structure into RAM using burst mode command 
			EXP_GET_FLIST */

		case 0x0250:
			buf->opcode=EXP_GET_FLIST;
			buf->usercode=0x0260;
#define cf buf->data.flist
			cf.start=0;
			cf.count=faces;
			cf.data=(_far FData *)fdata;
#undef cf
			break;

			/* Duplicate face structure in new object using burst mode 
			command EXP_PUT_FLIST. */
		case 0x0260:
			buf->opcode=EXP_PUT_FLIST;
			buf->usercode=0x0270;
#define cf buf->data.flist
			cf.start=0;
			cf.count=faces;
			cf.data=(_far FData *)fdata;
#undef cf
			break;

			/* Modify vertices working with copy of vertex data in 
			burst mode buffer vdata */
		case 0x0270:
			for(ix=0; ix<verts; ++ix)
				{
				if(state.type==ABSOLUTE)
					{
					switch(state.axis)
						{
						case XY:
							abswarp(&vdata[ix].x,&vdata[ix].y,&vdata[ix].z,centerx,centery);
							break;
						case XZ:
							abswarp(&vdata[ix].x,&vdata[ix].z,&vdata[ix].y,centerx,centerz);
							break;
						case YZ:
							abswarp(&vdata[ix].y,&vdata[ix].z,&vdata[ix].x,centery,centerz);
							break;
						case XYZ:
							abswarp(&vdata[ix].x,&vdata[ix].y,&vdata[ix].z,centerx,centery);
							abswarp(&vdata[ix].x,&vdata[ix].z,&vdata[ix].y,centerx,centerz);
							abswarp(&vdata[ix].y,&vdata[ix].z,&vdata[ix].x,centery,centerz);
							break;
						}
					}
				else
					relwarp(&vdata[ix].x,&vdata[ix].y,&vdata[ix].z,centerx,centery,centerz);
				}

			buf->opcode=EXP_PUT_VLIST;
			buf->usercode=0x0280;
#define cv buf->data.vlist
			cv.start=0;
			cv.count=verts;
			cv.data=(_far VData *)vdata;
#undef cv
			break;

		case 0x0280:
			create_it:
			buf->opcode=EXP_CREATE_OBJ;
			buf->usercode=0x0300;
#define ob buf->data.object
			sprintf(ob.name,"%s%03d",state.prefix,morphnum);
			if(morphnum>0 && state.hide)
				ob.flags=EXP_OBJ_HIDDEN;
			else
				ob.flags=0;
#undef ob
			break;

		case 0x0300:
			morphnum++;
			if(morphnum<state.morphs)
				goto morph_loop;

			buf->opcode=buf->usercode=EXP_TERMINATE;
			buf->status=1;
			break;

		default:
			terminate:
			buf->opcode=buf->usercode=EXP_TERMINATE;
			buf->status=0;
			break;
		}
	}

void ClientTerminate(void) { 
	/* free any data structures, etc. (burst mode buffers) */
	if(vdata)
		free(vdata);
	vdata=NULL;
	if(fdata)
		free(fdata);
	fdata=NULL;
	}

DlgEntry *ClientDialog(int n) { 
	return(&cdialog[n]); 
	}

/* Wave the point based on parameters -- absolute mode */

abswarp(float *x,float *y,float *z,float cx,float cy)
{
float dist,dx,dy;

dx= *x-cx;
dy= *y-cy;
dist=sqrt(dx*dx+dy*dy);
*z+=sin((dist/state.wavelength+(float)morphnum/(float)state.period)*PI_360) *
	state.amp;
}

/* Wave the point based on parameters -- relative mode */

relwarp(float *x,float *y,float *z,float cx,float cy,float cz)
{
float dist,dx,dy,dz,nx,ny,nz,displacement;

dx= *x-cx;
dy= *y-cy;
dz= *z-cz;
dist=sqrt(dx*dx+dy*dy+dz*dz);
if(dist==0.0)
 dist=0.0000001;
nx=dx/dist;
ny=dy/dist;
nz=dz/dist;
displacement=sin((dist/state.wavelength+(float)morphnum/(float)state.period)*PI_360) *
	state.amp;
*x+=(nx*displacement);
*y+=(ny*displacement);
*z+=(nz*displacement);
}
