/*       3D Studio's              */
/*  Tornado Process ver 1.1       */
/*  by Tom Hudson 12/15/91        */
/* (Modifed for 2.01 burst mode.) */

/* AXP process raw client routine core code */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "axp.h"
#include "trig.h"

#define Rand01() ((float)(rand() & 0x7fff) / 32767.0)

/* Variable definitions */

#define PARTICLES 1
#define SIZE 2
#define SEED 3
#define INNER 4
#define OUTER 5
#define CHAOS 6
#define ROTATION 7

/************************/

/* Dialog description */

DlgEntry cdialog[]={
	0,"TITLE=\"Autodesk 3D Studio Tornado Generator\"",
	0,"TITLE=\"by Tom Hudson, Version 1.1\"",
	0,"TITLE=\"\"",
	PARTICLES,"LINT=\"# of Particles:\",1,9999,4",
	SIZE,"LFLOAT=\"Object/Particle size ratio:\",0.001,0.5,6",
	SEED,"INT=\"Random # Seed:\",5",
	INNER,"INT=\"Inner Orbit Period (Frames):\",5",
	OUTER,"INT=\"Outer Orbit Period (Frames):\",5",
	0,"TITLE= \"          Low                High\"",
	CHAOS,"SLIDER=\"Chaos:\",0,100,24",
	ROTATION,"RADIO=\"Rotation:\",\"CW\",\"CCW\"",
	0,"TITLE=\"\"",
	0,NULL
	};

/* Version test value */

#define VERSION 0x4981
	
typedef struct {
	ulong version;
	int particles;
	float size;
	int seed,inner,outer,chaos,rotation;
	} State;

/* the "state" struct MUST start with a "ulong" which is the version#,
	to prevent using data from old versions of this program.
	This verification is performed automatically. */

static State init_state = { VERSION,100,0.02,12345,5,15,0,0 };
static State state = { VERSION,100,0.02,12345,5,15,0,0 };

/* Variables */

/* Declare burst mode arrays */
static VData *vdata=NULL;
static FData *fdata=NULL;

static int frame,particle,overts,burstmode,creverts,crefaces;
static float sizefac,partx,party,partz,px,py,pz,radius,orbit_time,position;
static float sinfac,cosfac,xfac,yfac,zfac,fchaos,forder,taper,ybasis,ylimit;
static int dialog_ix,verts,faces,first,ix,onedges,offedges,edges,totlights,edge;
static int vertex,face,base;
static float minx,miny,minz,maxx,maxy,maxz,dx,dy,dz;
static unsigned char material;

/*----------------------------------------------------------------*/

void ClientSetStateVar(int id, void *ptr) {
	OVL o;
	ulong *ul;

	ul=(ulong *)ptr;
	o.ul = *ul;
	switch(id) {
		case PARTICLES: state.particles=o.i; break;
		case SIZE: state.size=o.f; break;
		case SEED: state.seed=o.i; break;
		case INNER: state.inner=o.i; break;
		case OUTER: state.outer=o.i; break;
		case CHAOS: state.chaos=o.i; break;
		case ROTATION: state.rotation=o.i; break;
		}
	}

ulong ClientGetStateVar(int id) {
	OVL o;
	switch(id) {
		case PARTICLES: o.i=state.particles; break;
		case SIZE: o.f=state.size; break;
		case SEED: o.i=state.seed; break;
		case INNER: o.i=state.inner; break;
		case OUTER: o.i=state.outer; break;
		case CHAOS: o.i=state.chaos; break;
		case ROTATION: o.i=state.rotation; break;
		}
	return(o.ul);
	}

int ClientVarSize(int id)
	{
	switch(id)
		{
		default:
			return(1);
		}
	}

char  *ClientGetState(int *size) {
	*size = sizeof(State);
	return((char *)&state);
	}

void ClientResetState() {   
	state = init_state; 
	}

/* Client process starts here */

void ClientStartup(EXPbuf *buf) {
#define ob buf->data.object
	overts=ob.verts;    /* Get existing object's count for minmaxing */
	frame=ob.curfield;
#undef ob
	sizefac=state.size/2.0;

	/* Randomize for this object */
	srand(state.seed);

	/* Set burstmode flag based on 3D Studio version */
	burstmode=(studio_version()!=200) ? 1:0;

	/* If 3D Studio version is newer than 2.0, we can find the
	extents of our AXP stand using the faster EXP_OBJ_EXTENTS */
	if(burstmode)
		{
		/* Find object's extents */
		buf->opcode=EXP_OBJ_EXTENTS;    /* Studio does this */
		buf->usercode=0x0220;       /* Do this next time */
		buf->status=1;
		}
	else    /* Do it the Version 2.0 way... */
		{
		first=1;
		ix=0;

#define cv buf->data.vert
		cv.vertex=ix;
		buf->opcode=EXP_GET_VERTEX; /* Studio does this */
		buf->usercode=0x0200;       /* Do this next time */
		buf->status=1;
#undef cv
		}
	}

/* User routines -- use user codes to process your data */

void ClientUserCode(EXPbuf *buf)
	{
	switch(buf->usercode)
		{
		case 0x0200:    /* Extents */
#define cv buf->data.vert
			if(first)
				{
				minx=maxx=cv.x;
				miny=maxy=cv.y;
				minz=maxz=cv.z;
				first=0;
				}
			else
				{
				if(cv.x<minx)
					minx=cv.x;
				else
				if(cv.x>maxx)
					maxx=cv.x;
				if(cv.y<miny)
					miny=cv.y;
				else
				if(cv.y>maxy)
					maxy=cv.y;
				if(cv.z<minz)
					minz=cv.z;
				else
				if(cv.z>maxz)
					maxz=cv.z;
				}
			ix++;
			if(ix<overts)
				{
				cv.vertex=ix;
				buf->opcode=EXP_GET_VERTEX; /* Studio does this */
				buf->usercode=0x0200;       /* Do this next time */
				buf->status=1;
				break;
				}
#undef cv
			goto calc_deltas;

		case 0x0220:    /* Extents */
#define ex buf->data.exts
			minx=ex.minx;
			maxx=ex.maxx;
			miny=ex.miny;
			maxy=ex.maxy;
			minz=ex.minz;
			maxz=ex.maxz;
#undef ex
			calc_deltas:
			dx=maxx-minx;
			dy=maxy-miny;
			dz=maxz-minz;

			/* Find object material */

			buf->opcode=EXP_GET_FACE;
			buf->usercode=0x0230;
#define cf buf->data.face
			cf.face=0;
			break;
		case 0x0230:
			material=cf.material;
#undef cf
			/* Create the necessary number of verts & faces */

			/* These vert and face counts determine size of burst mode 
			data arrays */

			buf->opcode=EXP_READY_OBJ;
			buf->usercode=0x0240;
			buf->status=1;
#define ob buf->data.object
			ob.verts=creverts=6*state.particles;
			ob.faces=crefaces=8*state.particles;
			ob.tverts=0;
#undef ob
			break;
		case 0x0240:

			/* Allocate RAM for object work buffers for burst mode */

			if(studio_version()!=200)
				{
				burstmode=1;
				if((vdata=(VData *)malloc(creverts*sizeof(VData)))==NULL)
					{
					noram:
					buf->opcode=EXP_CONTINUE;
					buf->usercode=EXP_TERMINATE;
					strcpy(buf->data.string,"No RAM for object workspace");
					return;
					}
			   /* if((fdata=(FData *)malloc(crefaces*sizeof(FData)))==NULL) */
					if(1)
					goto noram;

				}
			else
				burstmode=0;
			vertex=face=0;
			if(buf->status==0)
				{
				terminate:
				buf->opcode=buf->usercode=EXP_TERMINATE;
				break;
				}
			particle=0;
			fchaos=(float)state.chaos/100.0;
			forder=1.0-fchaos;

			part_loop:
			if(particle>=state.particles)
				goto pvert_done;

			retry_y:
			ybasis=Rand01();
			party=1.0-ybasis;

			recalc:
			px=Rand01()-0.5;
			pz=Rand01()-0.5;
			radius=sqrt(px*px+pz*pz);
			if(radius>0.5)
				goto recalc;

			/* Scale down, making a cone, depending on chaos setting */

			taper=fchaos+(forder*party);
			if(Rand01()>=fchaos)
				{
				do_taper:
				px*=taper;
				pz*=taper;
				}
			else
				{
				taper+=(fchaos*2);
				if(taper>1.0)
					taper=1.0;
				goto do_taper;
				}

			orbit_time=(float)state.inner+(float)(state.outer-state.inner)*(radius/0.5);
			position=(float)frame/orbit_time*PI_360;
			if(state.rotation)
				position= -position;
			sinfac=sin(position);
			cosfac=cos(position);

			partx=(px*cosfac+pz*sinfac+0.5)*dx+minx;
			party=maxy-party*dy;
			partz=(pz*cosfac-px*sinfac+0.5)*dz+minz;
			xfac=sizefac*dx;
			yfac=sizefac*dy;
			zfac=sizefac*dz;

			/* Now create the appropriate vertices for this particle */

			/* if are using burst mode, we use store our newly created 
			vertex data in the vdata array */

			if(burstmode)
				{
				vdata[vertex].x=partx-xfac;
				vdata[vertex].y=party;
				vdata[vertex].z=partz;
				vdata[vertex].u=0.0;
				vdata[vertex++].v=0.0;

				vdata[vertex].x=partx;
				vdata[vertex].y=party-yfac;
				vdata[vertex].z=partz;
				vdata[vertex].u=0.0;
				vdata[vertex++].v=0.0;

				vdata[vertex].x=partx+xfac;
				vdata[vertex].y=party;
				vdata[vertex].z=partz;
				vdata[vertex].u=0.0;
				vdata[vertex++].v=0.0;

				vdata[vertex].x=partx;
				vdata[vertex].y=party+yfac;
				vdata[vertex].z=partz;
				vdata[vertex].u=0.0;
				vdata[vertex++].v=0.0;

				vdata[vertex].x=partx;
				vdata[vertex].y=party;
				vdata[vertex].z=partz-zfac;
				vdata[vertex].u=0.0;
				vdata[vertex++].v=0.0;

				vdata[vertex].x=partx;
				vdata[vertex].y=party;
				vdata[vertex].z=partz+zfac;
				vdata[vertex].u=0.0;
				vdata[vertex++].v=0.0;

			/* if we are using burst mode, the following "goto" sends 
			us back to the top of the particle creating loop (part_loop) */

				particle++;
				if(vertex<creverts)
					goto part_loop;

			/* if we are using burst mode, we use the EXP_PUT_VLIST command 
			to send our vertex data to 3D Studio */

				buf->opcode=EXP_PUT_VLIST;
				buf->usercode=0x0270;
#define cv buf->data.vlist
				cv.start=0;
				cv.count=creverts;
				cv.data=(_far VData *)vdata;
#undef cv
				break;
				}


			/* if we are NOT using burst mode, we skip from the top of 
			the particle creating loop (part_loop) to the following section 
			that uses the slower EXP_PUT_VERTEX command */

#define cv buf->data.vert
			buf->opcode=EXP_PUT_VERTEX;
			buf->usercode=0x0260;
			cv.x=partx-xfac;
			cv.y=party;
			cv.z=partz;
			cv.vertex=vertex++;
			break;

		case 0x0260:
			buf->opcode=EXP_PUT_VERTEX;
			buf->usercode=0x0261;
			cv.x=partx;
			cv.y=party-yfac;
			cv.z=partz;
			cv.vertex=vertex++;
			break;

		case 0x0261:
			buf->opcode=EXP_PUT_VERTEX;
			buf->usercode=0x0262;
			cv.x=partx+xfac;
			cv.y=party;
			cv.z=partz;
			cv.vertex=vertex++;
			break;

		case 0x0262:
			buf->opcode=EXP_PUT_VERTEX;
			buf->usercode=0x0263;
			cv.x=partx;
			cv.y=party+yfac;
			cv.z=partz;
			cv.vertex=vertex++;
			break;

		case 0x0263:
			buf->opcode=EXP_PUT_VERTEX;
			buf->usercode=0x0264;
			cv.x=partx;
			cv.y=party;
			cv.z=partz-zfac;
			cv.vertex=vertex++;
			break;

		case 0x0264:
			buf->opcode=EXP_PUT_VERTEX;
			buf->usercode=0x0265;
			cv.x=partx;
			cv.y=party;
			cv.z=partz+zfac;
			cv.vertex=vertex++;
			break;

		case 0x0265:

			/* if we are NOT using burst mode, the following "goto" sends 
			us back to the top of the particle creating loop (part_loop) */

			next_particle:
			particle++;
			goto part_loop;
#undef cv
			/* Now add the appropriate faces */

		case 0x0270:
			
			/* if are using burst mode, we use store our newly created 
			face data in the fdata array */
			
			pvert_done:
			ix=face=0;

			face_loop:
			base=ix*6;
			if(burstmode)
				{
				fdata[face].a=base;
				fdata[face].b=base+1;
				fdata[face].c=base+5;
				fdata[face].flags=FC_ABLINE | FC_BCLINE | FC_CALINE;
				fdata[face].sm_group=0;
				fdata[face++].material=material;

				fdata[face].a=base+1;
				fdata[face].b=base+2;
				fdata[face].c=base+5;
				fdata[face].flags=FC_ABLINE | FC_BCLINE | FC_CALINE;
				fdata[face].sm_group=0;
				fdata[face++].material=material;

				fdata[face].a=base+2;
				fdata[face].b=base+3;
				fdata[face].c=base+5;
				fdata[face].flags=FC_ABLINE | FC_BCLINE | FC_CALINE;
				fdata[face].sm_group=0;
				fdata[face++].material=material;

				fdata[face].a=base+3;
				fdata[face].b=base;
				fdata[face].c=base+5;
				fdata[face].flags=FC_ABLINE | FC_BCLINE | FC_CALINE;
				fdata[face].sm_group=0;
				fdata[face++].material=material;

				fdata[face].a=base;
				fdata[face].b=base+4;
				fdata[face].c=base+1;
				fdata[face].flags=FC_ABLINE | FC_BCLINE | FC_CALINE;
				fdata[face].sm_group=0;
				fdata[face++].material=material;

				fdata[face].a=base+1;
				fdata[face].b=base+4;
				fdata[face].c=base+2;
				fdata[face].flags=FC_ABLINE | FC_BCLINE | FC_CALINE;
				fdata[face].sm_group=0;
				fdata[face++].material=material;

				fdata[face].a=base+2;
				fdata[face].b=base+4;
				fdata[face].c=base+3;
				fdata[face].flags=FC_ABLINE | FC_BCLINE | FC_CALINE;
				fdata[face].sm_group=0;
				fdata[face++].material=material;

				fdata[face].a=base+3;
				fdata[face].b=base+4;
				fdata[face].c=base;
				fdata[face].flags=FC_ABLINE | FC_BCLINE | FC_CALINE;
				fdata[face].sm_group=0;
				fdata[face++].material=material;

			/* if we are using burst mode, the following "goto" sends us 
			back to the top of the face creating loop (face_loop) */

				ix++;
				if(ix<state.particles)
					goto face_loop;

			/* if we are using burst mode, we use the EXP_PUT_FLIST command 
			to send our face data to 3D Studio */

				buf->opcode=EXP_PUT_FLIST;
				buf->usercode=0x0350;
#define cf buf->data.flist
				cf.start=0;
				cf.count=crefaces;
				cf.data=(_far FData *)fdata;
#undef cf
				break;      
				}

			/* if we are NOT using burst mode, we skip from the top of the 
			face creating loop (face_loop) to the following section that 
			uses the slower EXP_PUT_FACE command */

#define cf buf->data.face
			buf->opcode=EXP_PUT_FACE;
			buf->usercode=0x0305;
			cf.a=base;
			cf.b=base+1;
			cf.c=base+5;
			cf.flags=FC_ABLINE | FC_BCLINE | FC_CALINE;
			cf.sm_group=0;
			cf.material=material;
			cf.face=face++;
			break;

		case 0x0305:
			buf->opcode=EXP_PUT_FACE;
			buf->usercode=0x0310;
			cf.a=base+1;
			cf.b=base+2;
			cf.c=base+5;
			cf.flags=FC_ABLINE | FC_BCLINE | FC_CALINE;
			cf.sm_group=0;
			cf.material=material;
			cf.face=face++;
			break;

		case 0x0310:
			buf->opcode=EXP_PUT_FACE;
			buf->usercode=0x0315;
			cf.a=base+2;
			cf.b=base+3;
			cf.c=base+5;
			cf.flags=FC_ABLINE | FC_BCLINE | FC_CALINE;
			cf.sm_group=0;
			cf.material=material;
			cf.face=face++;
			break;

		case 0x0315:
			buf->opcode=EXP_PUT_FACE;
			buf->usercode=0x0320;
			cf.a=base+3;
			cf.b=base;
			cf.c=base+5;
			cf.flags=FC_ABLINE | FC_BCLINE | FC_CALINE;
			cf.sm_group=0;
			cf.material=material;
			cf.face=face++;
			break;

		case 0x0320:
			buf->opcode=EXP_PUT_FACE;
			buf->usercode=0x0325;
			cf.a=base;
			cf.b=base+4;
			cf.c=base+1;
			cf.flags=FC_ABLINE | FC_BCLINE | FC_CALINE;
			cf.sm_group=0;
			cf.material=material;
			cf.face=face++;
			break;

		case 0x0325:
			buf->opcode=EXP_PUT_FACE;
			buf->usercode=0x0330;
			cf.a=base+1;
			cf.b=base+4;
			cf.c=base+2;
			cf.flags=FC_ABLINE | FC_BCLINE | FC_CALINE;
			cf.sm_group=0;
			cf.material=material;
			cf.face=face++;
			break;

		case 0x0330:
			buf->opcode=EXP_PUT_FACE;
			buf->usercode=0x0335;
			cf.a=base+2;
			cf.b=base+4;
			cf.c=base+3;
			cf.flags=FC_ABLINE | FC_BCLINE | FC_CALINE;
			cf.sm_group=0;
			cf.material=material;
			cf.face=face++;
			break;

		case 0x0335:
			buf->opcode=EXP_PUT_FACE;
			buf->usercode=0x0340;
			cf.a=base+3;
			cf.b=base+4;
			cf.c=base;
			cf.flags=FC_ABLINE | FC_BCLINE | FC_CALINE;
			cf.sm_group=0;
			cf.material=material;
			cf.face=face++;
			break;

		case 0x0340:

			/* if we are NOT using burst mode, the following "goto" sends 
			us back to the top of the face creating loop (face_loop) */

			ix++;
			if(ix>=state.particles)
				goto createit;
			goto face_loop;

		case 0x0350:
			createit:
			buf->opcode=EXP_CREATE_OBJ;
			buf->usercode=EXP_TERMINATE;
			buf->status=1;
			break;

		default:
		error("Invalid user opcode",NULL,NULL,NULL,NULL);
		buf->opcode=buf->usercode=EXP_TERMINATE;
		buf->status=0;
		}
	}

/* check if any burst mode buffers were allocated and release that memory */

void ClientTerminate(void) { 
	/* free any data structures, etc. */
	if(vdata)
		free(vdata);
	vdata=NULL;
	if(fdata)
		free(fdata);
	fdata=NULL;
	}

DlgEntry *ClientDialog(int n) { 
	return(&cdialog[n]); 
	}

char *ClientName(void)
	{
	return(".AXP");     /* Put your process name here */
	}
