/*       3D Studio's              */
/*  Cross-hair object locator     */
/*  by Tom Hudson 8/17/92         */
/*          ver. 1.0              */
/* See readme.doc for information regarding the required         */
/* OUTPUT-RENDER-COORDS = YES 3DS.SET statement so this program  */
/* can function.  (Note that this program will only work         */
/* with versions of 3D Studio 2.01 or greater.)                  */

/* IXP process raw client routine core code */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "ixp.h"
#include "trig.h"

#define Rand01() ((float)(rand() & 0x7fff) / 32767.0)

typedef float FLOAT;
typedef FLOAT TRow[3];
typedef TRow TMatrix[4];

extern void split_fn(char *path,char *file,char *pf);
extern void split_fext(char *fe,char *f,char *e);
extern ch_replace(char *string,char oldch,char newch); 
extern void xform_point(float x,float y,float z,
	float *rx,float *ry,float *rz);
extern void frect(float x, float y, float w, float h,int brite);
extern int do_xhair(float minx,float maxx,float miny,float maxy);

#define RelScaleX(tm, s) RelScaleRow(tm[0],s) 
#define RelScaleY(tm, s) RelScaleRow(tm[1],s) 
#define RelScaleZ(tm, s) RelScaleRow(tm[2],s)

#define EVEN_LINES 1
#define ODD_LINES 2
#define ALL_LINES 3

typedef struct
{
float pos;
float size;
float density;
} Flare;
static Flare flarelist[21];

/* Frame params */

static float dev_aspect,ov_zoom,view_hangle,view_vangle,view_tilt;
static int dev_width,dev_height,view_width,view_type,proj_type,field_type;
static float xmat[4][3],i_xmat[4][3],basemat[4][3],refmat[4][3];
static float heading,pitch,bank,camzfac,camw2,camh2,vpt_x,vpt_y,vpt_z;
static float scale,yscale,camdist;
static float dimmest,brightest;
static float ident_mat[12]={1.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0};
static FrameInfo *fi;
static FrameParm *fp;
static FieldInfo *fld;

/* Star database code from stars.ixp */

typedef struct
{
float x,y,z,fmag;
unsigned char mag;
} Star;

/* Variables */

static int dim,bright,logarithmic;
static char infile[13];

/* Variable definitions */

#define HAIRSIZE 1
#define COLOR 2
#define OBJECT 3

/************************/

/* Dialog description */

DlgEntry cdialog[]={
	0,"TITLE=\"Autodesk 3D Studio Crosshair Generator\"",
	0,"TITLE=\"Version 1.0\"",
	0,"TITLE=\"Copyright 1993 Tom Hudson.  From Yost Group, Inc.\"",
	0,"TITLE=\"This is a Public Domain IPAS Routine.  Not for Resale\"",
	0,"TITLE=\"\"",
	HAIRSIZE,"SLIDER=\"Crosshair size:\",3,40,20",
	COLOR,"COLOR=\"Crosshair color\"",
	OBJECT,"STRING=\"Object to target:\",11,11",
	0,"TITLE=\"\"",
	0,NULL
	};

/* Version test value */

#define VERSION 0xE05F
	
typedef struct {
	ulong version;
	int hairsize;
	Col24 color;
	char object[11],filler[1];
	} State;

/* the "state" struct MUST start with a "ulong" which is the version#,
	to prevent using data from old versions of this program.
	This verification is performed automatically. */

static State init_state = { VERSION,20,{255,255,255},"Object01","" };
static State state = { VERSION,20,{255,255,255},"Object01","" };

/*----------------------------------------------------------------*/

int ClientProcImage(FrameParm *f,FrameInfo *i,FieldInfo *d,int state,Report *rpt) {
	fi=i;
	fp=f;
	fld=d;

	if(studio_version()==200)
		{
		error("XHAIR.IXP requires 3D Studio","Version 2.01 or higher",NULL,NULL,NULL);
		return(0);
		}

	dev_width=fi->dev_width;
	dev_height=fi->dev_height;
	dev_aspect=fi->dev_aspect;
	ov_zoom=fi->ov_zoom;

	/*
	Initialize the output image area:
	Copy the required number of pixels from the input map to the output map.
	*/

	far_to_far(fp->outpix,fp->inpix,dev_width*dev_height*3);

	if(fld->using_fields)
		{

		/* First, even scan lines */

		process_image((fld->field_order) ? ODD_LINES:EVEN_LINES);

		/* Next, odd scan lines */

		process_image((fld->field_order) ? EVEN_LINES:ODD_LINES);
		}
	else		/* Frame mode */
		{
		process_image(ALL_LINES);
		}

	return(1);
	}

void ClientSetStateVar(int id, void *ptr) {
 	OVL o;
	ulong *ul;
	char *s;

	ul=(ulong *)ptr;
	s=(char *)ptr;
	o.ul = *ul;
	switch(id) {
		case HAIRSIZE: state.hairsize=o.i; break;
		case COLOR:
			state.color.r = o.c.r;
			state.color.g = o.c.g; 
			state.color.b = o.c.b; 
			break;
		case OBJECT: strcpy(state.object,s); break;
		}
	}

ulong ClientGetStateVar(int id) {
	OVL o;
	switch(id) {
		case HAIRSIZE: o.i=state.hairsize; break;
		case COLOR:
			o.c.r = state.color.r;
			o.c.g = state.color.g;
			o.c.b = state.color.b;
			break;
		case OBJECT: o.s=state.object; break;
		}
	return(o.ul);
	}

int ClientVarSize(int id) {
	switch(id) {
		case OBJECT:
			return(3);
		default:
			return(1);
		}
	}

char  *ClientGetState(int *size) {
	*size = sizeof(State);
	return((char *)&state);
	}

void ClientResetState() {	
	state = init_state;	
	}

void ClientStartup(void) {
	/* allocate data structures, compute lookup tables, etc */
	}

void ClientTerminate(void) { 
	/* free any data structures, etc. */
	free_alloced();
	}

DlgEntry *ClientDialog(int n) {	
	return(&cdialog[n]); 
	}

free_alloced()
{
}

process_image(int type)
{
float minx,miny,maxx,maxy;
int line,result;
char *fname,object[15];
FILE *istream;

field_type=type;

switch(field_type)
 {
 case EVEN_LINES:
 case ALL_LINES:
  fname="3dseven.loc";
  break;
 case ODD_LINES:
  fname="3dsodd.loc";
  break;
 }

/* Read file -- if object we want isn't listed, exit */

if((istream=fopen(fname,"r"))==NULL)
 {
 error("Can't open .LOC file:",fname,"Please be sure to set the 3DS.SET","OUTPUT-RENDER-COORDS=YES",NULL);
 return;
 }

loc_loop:
result=fscanf(istream,"%12s %f %f %f %f\n",object,&minx,&maxx,&miny,&maxy);
if(result==EOF)
 {
 loc_done:
 fclose(istream);
 return;
 }
if(result!=5)
 {
 error("Invalid .LOC file:",fname,NULL,NULL,NULL);
 return;
 }
if(name_match(state.object,object))
 do_xhair(minx,maxx,miny,maxy);
goto loc_loop;
}

int
do_xhair(float minx,float maxx,float miny,float maxy)
{
float ctrx,ctry,width,height;

ctrx=(minx+maxx)/2.0;
ctry=(miny+maxy)/2.0;

/* If object center offscreen, forget it */

if(ctrx<0 || ctrx>=dev_width || ctry<0 || ctry>=dev_height)
 return(0);

width=state.hairsize;
height=state.hairsize*dev_aspect;

/* Plot the crosshair */

frect(ctrx,ctry,width,1.5,255);
frect(ctrx,ctry,1.5,height,255);
return(1);
}

void plot(int x, int y, uchar v )
{
int pixel,iv;

if((field_type==EVEN_LINES && (y&1)) || (field_type==ODD_LINES && ((y&1)==0)))
 return;

if(x<0 || y<0 || x>=dev_width || y>=dev_height)
 return;

pixel=dev_width*y+x;
iv=255-v;
fp->outpix[pixel].r=((fp->outpix[pixel].r*iv)+(state.color.r*v))/255;
fp->outpix[pixel].g=((fp->outpix[pixel].g*iv)+(state.color.g*v))/255;
fp->outpix[pixel].b=((fp->outpix[pixel].b*iv)+(state.color.b*v))/255;
}


/*------------------------------------------------------------
Dan's Handy-Dandy Floating-Point Rectangle Plotter (DHDFPRP) ...
	
	Renders an anti-aliased floating point rectangle on pixel grid.
	x,y,w,h are floating point screen coords. 

	Tom's modified version: x&y are the center of the rectangle.

	(Note: This could be sped up using fixed point.)

-------------------------------------------------------------*/

#define frac(z) fmod(z,1.0)
#define whole(z) (z-fmod(z,1.0))

void frect(float x, float y, float w, float h,int brite) {
	int xa,xb,ya,yb,i,j;
	float x2,y2,lfrac,rfrac,tfrac,bfrac,xfrac,yfrac;
	x -= (w/2.0);
	y -= (h/2.0);
	x2 = x + w;
	y2 = y + h;
  	xa = (int)x;
	xb = (int)x2;
  	ya = (int)y;
	yb = (int)y2;
	lfrac = 1.0-frac(x);	
	rfrac = frac(x2);
	tfrac = 1.0-frac(y);	
	bfrac = frac(y2);
	if (xa==xb) {
		xfrac = lfrac+rfrac-1.0;
		if (ya==yb) { 
			/* --- only one pixel effected */
			plot(xa,ya,xfrac*(tfrac+bfrac-1.0)*brite);	
			}
		else {  
			/* --- column of pixels */
			plot(xa,ya,xfrac*tfrac*brite);
			for (j=ya+1; j<yb; j++) plot(xa,j,xfrac*brite);
			plot(xa,yb,xfrac*bfrac*brite);
			}
		}
	else {
		if (ya==yb) { 
			/* --- row of pixels */
			yfrac = tfrac+bfrac-1.0;
			plot(xa,ya,yfrac*lfrac*brite);
			for (i=xa+1; i<xb; i++) plot(i,ya,yfrac*brite);
			plot(xb,ya,yfrac*rfrac*brite);
			}
		else {  
			/* general case */

			/* --- top row */
			plot(xa,ya,tfrac*lfrac*brite);
			for (i=xa+1; i<xb; i++) plot(i,ya,tfrac*brite);
			plot(xb,ya,tfrac*rfrac*brite);

			/* --- middle (whole pixels ) */
			for (j=ya+1; j<yb; j++) {
				plot(xa,j,lfrac*brite);
				for (i=xa+1; i<xb; i++) plot(i,j,brite);
				plot(xb,j,rfrac*brite);
				}

			/* --- bottom row */
			plot(xa,yb,bfrac*lfrac*brite);
			for (i=xa+1; i<xb; i++) plot(i,yb,bfrac*brite);
			plot(xb,yb,bfrac*rfrac*brite);
			}
		}
	}

char *ClientName(void)
{
return("XHAIR.IXP");
}

/* Wildcard match */

name_match(char *pat,char *string)
{
while(1)
 {
 switch(*pat)
  {
  case '*':
   return(1);
  case '?':
   if(*string==0)
    return(0);
   break;
  default:
   if(*pat!=*string)
    return(0);
   if(*pat==0)
    return(1);
   break;
  }
 pat++;
 string++;
 }
}


